; File:		updi-tiny3216\main.asm
; Device:	ATtiny3216
; Created:	09.06.2024 21:52:27
; Version:	2024-06-18
; Author:	Johannes Fechner

.nolist
.include <tn3216def.inc>
.include "..\..\..\..\avr-asm-lib\macros.asm"
.list

; == Hardware definitions ==
; === CPU clock frequency ===
.equ F_CPU = 18_432_000 ; Hz

; === Hardware connections ===
; USART0.TxD output:
.equ TXD_VDIR = VPORTB_DIR
.equ TXD_bp = 2
; Target UPDI (connected via 1kOhm resistor to UPDI of target device):
.equ TARGET_UPDI_INTFLAGS = PORTC_INTFLAGS
.equ TARGET_UPDI_PINCTRL = PORTC_PIN0CTRL
.equ TARGET_UPDI_PORT_VECT = PORTC_PORT_vect
.equ TARGET_UPDI_VDIR = VPORTC_DIR
.equ TARGET_UPDI_VIN = VPORTC_IN
.equ TARGET_UPDI_VOUT = VPORTC_OUT
.equ TARGET_UPDI_bp = 0
; Target UPDI Rx LED:
.equ LED_UPDI_RX_VDIR = VPORTB_DIR
.equ LED_UPDI_RX_VIN = VPORTB_IN
.equ LED_UPDI_RX_VOUT = VPORTB_OUT
.equ LED_UPDI_RX_bp = 0
; Target UPDI Tx LED:
.equ LED_UPDI_TX_VDIR = VPORTB_DIR
.equ LED_UPDI_TX_VOUT = VPORTB_OUT
.equ LED_UPDI_TX_bp = 1

; == PC UART definitions ==
.equ UART_BPS = 115_200 ; bits/s

; == UPDI protocol definitions ==
.equ UPDI_BPS = 9_600		; UPDI UART baud rate in bits/s.
.equ UPDI_UART_BUFLEN = 16	; UPDI UART tx/rx buffer length in bytes.

; == Register name definitions ==
.def reg00 = r2		; Contains always 0x00.

; === Temporary registers ===
.def temp0 = r16
.def temp1 = r17
.def temp2 = r18
.def temp3 = r19
.def temp4 = r20
.def temp5 = r21
.def tempL = r24
.def tempH = r25

; === Main registers ===
.def updiFlags = r3				; UPDI status bits and flags.
.equ UPDI_FLAGS_BUSY_bp = 0		; UPDI transmission or reception in progress.
.equ UPDI_FLAGS_TXC_bp = 1		; UPDI transmission complete.
.equ UPDI_FLAGS_RXC_bp = 2		; UPDI Frame Received flag.
.equ UPDI_FLAGS_TIMEOUT_bp = 3	; Timeout occurred while waiting for reception.
.equ UPDI_FLAGS_BUFOVF_bp = 4	; Buffer overflow while receiving.

; == RAM usage ==
.dseg
.org SRAM_START

updiUartBitCntr_ram:		.byte   1
; == 0: during start bit.
; == 1: during data LSb.
; ...
; == 8: during data MSb.
; == 9: during (even) parity bit.
; ==10: during 1st stop bit.
; ==11: during 2nd stop bit.
updiUartBitMask_ram:		.byte   1
updiUartDataBuf_ram:		.byte   UPDI_UART_BUFLEN
updiUartDataCount_ram:		.byte	1
updiUartDataIndex_ram:		.byte	1
updiUartParityCntr_ram:		.byte   1
updiUartTimeoutCntr_ram:	.byte   1

writeFuseAddrL_ram:			.byte   1
writeFuseAddrH_ram:			.byte   1
writeFuseData_ram:			.byte   1

.cseg
.org 0
	rjmp	reset

; == Interrupt vectors ==
.org TARGET_UPDI_PORT_VECT
	rjmp	updi_fallingEdgeIsr
.org TCA0_OVF_vect
	rjmp	updi_timerIsr
	
; == Zero-terminated strings ==
.org INT_VECTORS_SIZE
szMenu: .db "(1) Read SIB", '\r', '\n',\
	"(2) Read UPDI register", '\r', '\n',\
	"(3) Read byte from UPDI address space", '\r', '\n',\
	"(4) Write fuse byte", '\r', '\n',\
	"(5) Chip erase", '\r', '\n', 0, 0
szAddress: .db "Address: ", 0
szData: .db " Data: ", 0
szConfirm: .db " Confirm (y/n): ", 0, 0
szEnteringCEK: .db "Entering Chip Erase key...", '\r', '\n', 0, 0
szEnteringNVMPK: .db "Entering NVM Programming key...", '\r', '\n', 0
szAsiKeyStatus: .db "ASI KEY STATUS: ", 0, 0
szReset: .db "Issuing System Reset...", '\r', '\n', 0
szAsiSysStatus: .db "ASI SYS STATUS: ", 0, 0
szNvmctrlStatus: .db "NVMCTRL_STATUS: ", 0, 0
szErrBufOvf: .db "Error: Buffer overflow while receiving.", '\r', '\n', '\r', '\n', 0
szErrTimeout: .db "Error: Timeout while waiting for UPDI answer.", '\r', '\n', '\r', '\n', 0
szErrKey: .db "Error: ASI Key submitting failed.", '\r', '\n', '\r', '\n', 0
szLineBreak: .db '\r', '\n', 0, 0

; == External routines ==
.include "..\..\..\..\avr-asm-lib\updi.asm"

; == Execution entry point after reset ==
reset:
; == Set clock ==
; Set to external clock:
	ldi		temp0, CPU_CCP_IOREG_gc
	xout	CPU_CCP, temp0
	ldi		temp0, CLKCTRL_CLKSEL_EXTCLK_gc
	xout	CLKCTRL_MCLKCTRLA, temp0
; Disable clock prescaler:
	ldi		temp0, CPU_CCP_IOREG_gc
	xout	CPU_CCP, temp0
	ldi		temp0, 0
	xout	CLKCTRL_MCLKCTRLB, temp0

; == Initialize registers ==
	clr		reg00
	clr		updiFlags

; == Initialize UART ==
.equ BAUD_REG_VAL = ROUND(4.0*F_CPU/(1.0*UART_BPS))
	ldi		temp0, low(BAUD_REG_VAL)
	xout	USART0_BAUDL, temp0
	ldi		temp0, high(BAUD_REG_VAL)
	xout	USART0_BAUDH, temp0
; USART0_CTRLC = $03 (reset value): asynchronous, 8N1.
	ldi		temp0, (1<<USART_TXEN_bp) | (1<<USART_RXEN_bp)
	xout	USART0_CTRLB, temp0

; == Initialize TCA0 for UPDI Soft-UART ==
; Enable overflow interrupt:
	ldi		temp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTCTRL, temp0

; == Initialize I/O ==
; Configure TxD pin as output:
	sbi		TXD_VDIR, TXD_bp
; Configure Rx/Tx LEDs as outputs:
	sbi		LED_UPDI_RX_VDIR, LED_UPDI_RX_bp
	sbi		LED_UPDI_TX_VDIR, LED_UPDI_TX_bp
; Enable pull-up at Target UPDI pin:
	ldi		temp0, PORT_PULLUPEN_bm
	xout	TARGET_UPDI_PINCTRL, temp0

; == Enable interrupts ==
	sei
	
loop:
; Print menu to PC UART:
	ldiz	2*szMenu
	rcall	uart_putSz
; Wait for PC UART reception:
	rcall	uart_getChar
	cpi		temp0, '1'
	brne	loop_2
	rjmp	loop_readSib
loop_2:
	cpi		temp0, '2'
	brne	loop_3
	rjmp	loop_readUpdiReg
loop_3:
	cpi		temp0, '3'
	brne	loop_4
	rjmp	loop_readByte
loop_4:
	cpi		temp0, '4'
	brne	loop_5
	rjmp	loop_writeFuse
loop_5:
	cpi		temp0, '5'
	brne	loop_default
	rjmp	loop_chipErase
loop_default:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop

loop_readSib:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
	/*xin		temp1, TCA0_SINGLE_CNT
	xin		temp0, TCA0_SINGLE_CNT+1
	rcall	uart_putHexByte
	mov		temp0, temp1
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz*/
; Send Key Get SIB command:
	ldi		temp0, UPDI_SYNCH
	sts		updiUartDataBuf_ram, temp0
	ldi		temp0, 0b111_00_1_01
	sts		updiUartDataBuf_ram+1, temp0
	ldi		temp0, 2	; Count of bytes to be sent.
	sts		updiUartDataCount_ram, temp0
	rcall	updi_tx
; Wait for transmission to finish:
loop_readSib_waitTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_readSib_waitTx
; Wait until 16 bytes were received:
loop_readSib_waitRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		temp0, updiUartDataIndex_ram
	cpi		temp0, 16
	brlo	loop_readSib_waitRx
; updiUartDataBuf_ram now contains the received bytes.
	clr		temp1
    ldiz	updiUartDataBuf_ram
loop_readSib_print:
	ld		temp0, Z+
    rcall   uart_putChar
	inc		temp1
	cpi		temp1, 16
	brlo	loop_readSib_print
	ldiz	2*szLineBreak
	rcall	uart_putSz
    ldiz    2*szLineBreak
    rcall   uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_readUpdiReg:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Get address from user:
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte
; Enable Target UPDI:
	rcall	updi_enable
; Send LDCS instruction:
	andi	temp0, 0x0F
	rcall	sendLdcs
; Wait for answer bytes from Target UPDI:
loop_readUpdiReg_wait:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_readUpdiReg_wait
; updiUartData_ram now contains the received byte.
    lds     temp0, updiUartDataBuf_ram
	ldiz	2*szData
	rcall	uart_putSz
    rcall   uart_putHexByte
    ldiz    2*szLineBreak
    rcall   uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_readByte:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Get address from user:
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte	; Address MSB.
	mov		temp1, temp0
	rcall	uart_getHexByte	; Address LSB.
; Enable Target UPDI:
	rcall	updi_enable
; Send LDS instruction:
	rcall	sendLds
; Read answer byte from Target UPDI:
loop_readByte_wait:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_readByte_wait
; updiUartData_ram now contains the received byte.
    lds     temp0, updiUartDataBuf_ram
	ldiz	2*szData
	rcall	uart_putSz
    rcall   uart_putHexByte
    ldiz    2*szLineBreak
    rcall   uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop
	
loop_writeFuse:
; == Echo the entered character ==
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; == Get address from user ==
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte
	sts		writeFuseAddrH_ram, temp0
	rcall	uart_getHexByte
	sts		writeFuseAddrL_ram, temp0
; == Get data from user ==
	ldiz	2*szData
	rcall	uart_putSz
	rcall	uart_getHexByte
	sts		writeFuseData_ram, temp0
	ldiz	2*szConfirm
	rcall	uart_putSz
	rcall	uart_getChar
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
	cpi		temp0, 'y'
	breq	loop_writeFuse_confirmed
; Operation cancelled by user.
	rjmp	loop
loop_writeFuse_confirmed:
; == Enable Target UPDI ==
	rcall	updi_enable
	rcall	updi_idle
; == Send NVMPROG key ==
	ldiz	2*szEnteringNVMPK
	rcall	uart_putSz
	rcall	sendNvmProgKey
; == Confirm success by reading ASI Key Status Register ==
	ldi		temp0, 0x07		; Address of ASI_KEY_STATUS.
	rcall	sendLdcs
; Read answer from target UPDI:
loop_writeFuse_waitKeyStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_writeFuse_waitKeyStatus
	ldiz	2*szAsiKeyStatus
	rcall	uart_putSz
	lds		temp0, updiUartDataBuf_ram
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	bst		temp0, 4		; NVMPROG bit.
	brts	loop_writeFuse_nvmKeySuccess
; NVMPROG key transmitting failed.
	ldiz	2*szErrKey
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop
loop_writeFuse_nvmKeySuccess:
; == Issue a System Reset to the target device ==
	ldiz	2*szReset
	rcall	uart_putSz
	rcall	systemReset
; == Wait until PROGSTART bit is 1 ==
loop_writeFuse_progStartLoop:
	ldi		temp0, 0x0B		; Address of ASI_SYS_STATUS.
	rcall	sendLdcs
; Read answer from target UPDI:
loop_writeFuse_waitSysStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_writeFuse_waitSysStatus
	lds		temp0, updiUartDataBuf_ram
	sbrs	temp0, 3		; Skip if PROGSTART bit is 1.
	rjmp	loop_writeFuse_progStartLoop
	ldiz	2*szAsiSysStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; == Do the NVM operation ==
; Write address LSB:
	ldi		temp0, low(NVMCTRL_ADDR)
	ldi		temp1, high(NVMCTRL_ADDR)
	lds		temp2, writeFuseAddrL_ram
	rcall	sendSts
; Write address MSB:
	ldi		temp0, low(NVMCTRL_ADDR+1)
	ldi		temp1, high(NVMCTRL_ADDR+1)
	lds		temp2, writeFuseAddrH_ram
	rcall	sendSts
; Write data:
	ldi		temp0, low(NVMCTRL_DATA)
	ldi		temp1, high(NVMCTRL_DATA)
	lds		temp2, writeFuseData_ram
	rcall	sendSts
; Issue Write Fuse command:
	ldi		temp0, low(NVMCTRL_CTRLA)
	ldi		temp1, high(NVMCTRL_CTRLA)
	ldi		temp2, NVMCTRL_CMD_FUSEWRITE_gc
	rcall	sendSts
; == Wait while NVMCTRL is busy ==
	rcall	waitForNvm
	sbrc	temp3, 0
	rjmp	loop_timeout
	ldiz	2*szNvmctrlStatus
	rcall	uart_putSz
	mov		temp0, temp2
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; == Issue a System Reset to the target device ==
	ldiz	2*szReset
	rcall	uart_putSz
	rcall	systemReset
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_chipErase:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send the Chip Erase key:
	ldiz	2*szEnteringCEK
	rcall	uart_putSz
	rcall	sendChipEraseKey
; Send the NVMPROG key:
/*	ldiz	2*szEnteringNVMPK
	rcall	uart_putSz
	rcall	sendNvmProgKey */
; == Confirm success by reading ASI Key Status Register ==
	ldi		temp0, 0x07		; Address of ASI_KEY_STATUS.
	rcall	sendLdcs
; Read answer from target UPDI:
loop_chipErase_waitKeyStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_chipErase_waitKeyStatus
	ldiz	2*szAsiKeyStatus
	rcall	uart_putSz
	lds		temp0, updiUartDataBuf_ram
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	bst		temp0, 3		; CHIPERASE bit.
	brts	loop_chipErase_keySuccess
; CHIPERASE key transmitting failed.
	ldiz	2*szErrKey
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop
loop_chipErase_keySuccess:
; == Issue a System Reset to the target device ==
	ldiz	2*szReset
	rcall	uart_putSz
loop_chipErase_waitReset:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	loop_chipErase_waitReset
	ldiz	updiUartDataBuf_ram
	ldi		temp0, UPDI_SYNCH
	st		Z+, temp0
	ldi		temp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	st		Z+, temp0
	ldi		temp0, 0x59					; Normal Reset signature.
	st		Z+, temp0
	ldi		temp0, 3					; Count of bytes to be sent.
	sts		updiUartDataCount_ram, temp0
	rcall	updi_tx
	rcall	updi_idle
loop_chipErase_clearReset:
	ldiz	updiUartDataBuf_ram
	ldi		temp0, UPDI_SYNCH
	st		Z+, temp0
	ldi		temp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	st		Z+, temp0
	ldi		temp0, 0x00					; Clear reset.
	st		Z+, temp0
	ldi		temp0, 3					; Count of bytes to be sent.
	sts		updiUartDataCount_ram, temp0
	rcall	updi_tx
	rcall	updi_idle
; Read LOCKSTATUS:
	ldi		temp0, 0x0B		; Address of ASI_SYS_STATUS.
	rcall	sendLdcs
; Read answer from target UPDI:
loop_chipErase_waitLockStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_chipErase_waitLockStatus
	lds		temp0, updiUartDataBuf_ram
	sbrc	temp0, 0		; Skip if LOCKSTATUS bit is 0.
	rjmp	loop_chipErase_clearReset
	ldiz	2*szAsiSysStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_timeout:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrTimeout
	rcall	uart_putSz
; Send BREAK to re-establish synchronization and disable UPDI:
	rcall	updi_break
	rcall	updi_disable
	rjmp	loop
	
; == Subroutines ==
; === Routine: Send character via UART ===
; Parameter: temp0.
uart_putChar:
	push	temp0
; Make sure the transmit data registers are empty:
uart_sendByte_wait:
	xin		temp0, USART0_STATUS
	sbrs	temp0, USART_DREIF_bp
	rjmp	uart_sendByte_wait
	pop		temp0
	xout	USART0_TXDATAL, temp0
	ret

; === Routine: Send string via UART ===
; Parameter: Z pointer, start address of the null-terminated string.
; String length is limited to 255 characters.
uart_putSz:
	push	temp0
	push	temp1
	clr		temp1
uart_putSz_loop:
	lpm		temp0, Z+
	tst		temp0
	breq	uart_putSz_end
	rcall	uart_putChar
	inc		temp1
	cpi		temp1, $FF
	breq	uart_putSz_end
	rjmp	uart_putSz_loop
uart_putSz_end:
	pop		temp1
	pop		temp0
	ret

; === Routine: Send byte in hexadecimal via UART ===
; Send 8-bit unsigned number as two hex digits.
; Parameter: temp0.
uart_putHexByte:
	push	temp0
	swap	temp0
	andi	temp0, 0x0F
	rcall	uart_putHexDigit
	pop		temp0
	push	temp0
	andi	temp0, 0x0F
	rcall	uart_putHexDigit
	pop		temp0
	ret
uart_putHexDigit:
	cpi		temp0, 10
	brlt	uart_putHexDigit_1
	subi	temp0, -('A'-'9'-1)
uart_putHexDigit_1:
	subi	temp0, -'0'
	rcall	uart_putChar
	ret

; === Wait for reception of character from UART ===
; Return: temp0, the received character.
uart_getChar:
	xin		temp0, USART0_STATUS
	sbrs	temp0, USART_RXCIF_bp
	rjmp	uart_getChar
	xin		temp0, USART0_RXDATAL
	ret
	
; === Convert ASCII hex digit character to its value ===
; Parameter:	temp0, the hex digit (ASCII) to be converted.
; Return:		temp1, $FF if given character is no hex digit.
hexChar2num:
	mov		temp1, temp0
; Convert small letters to capitals:
	sbrc	temp1, 6
	andi	temp1, ~(1<<5)
	cpi		temp1, '0'
	brlo	hexChar2num_err
	cpi		temp1, '9'+1
	brlo	hexChar2num_09
	cpi		temp1, 'A'
	brlo	hexChar2num_err
	cpi		temp1, 'F'+1
	brlo	hexChar2num_AF
hexChar2num_err:
	ldi		temp1, $FF
	rjmp	hexChar2num_end
hexChar2num_09:
	subi	temp1, '0'
	rjmp	hexChar2num_end
hexChar2num_AF:
	subi	temp1, 'A'-10
hexChar2num_end:
	ret

; == Read byte as two hex chars from UART ==
; Return:		temp0.
uart_getHexByte:
	push	temp1
	push	temp2
uart_getHexByte_get1stChar:
	rcall	uart_getChar
	rcall	hexChar2num
	cpi		temp1, $FF
	breq	uart_getHexByte_get1stChar
	rcall	uart_putChar				; Echo the first entered character.
	mov		temp2, temp1
uart_getHexByte_get2ndChar:
	rcall	uart_getChar
	rcall	hexChar2num
	cpi		temp1, $FF
	breq	uart_getHexByte_get2ndChar
	rcall	uart_putChar				; Echo the second entered character.
	swap	temp2
	or		temp1, temp2
	mov		temp0, temp1
	pop		temp2
	pop		temp1
	ret

; == Send the NVMPROG key ==
sendNvmProgKey:
	push	temp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendNvmProgKey_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendNvmProgKey_wait
	ldiz	updiUartDataBuf_ram
	ldi		temp0, UPDI_SYNCH
	st		Z+, temp0
	ldi		temp0, 0b111_00_0_00	; Send 64-bit key.
	st		Z+, temp0
	ldi		temp0, 0x20				; Key byte 0.
	st		Z+, temp0
	ldi		temp0, 0x67				; Key byte 1.
	st		Z+, temp0
	ldi		temp0, 0x6F				; Key byte 2.
	st		Z+, temp0
	ldi		temp0, 0x72				; Key byte 3.
	st		Z+, temp0
	ldi		temp0, 0x50				; Key byte 4.
	st		Z+, temp0
	ldi		temp0, 0x4D				; Key byte 5.
	st		Z+, temp0
	ldi		temp0, 0x56				; Key byte 6.
	st		Z+, temp0
	ldi		temp0, 0x4E				; Key byte 7.
	st		Z+, temp0
	ldi		temp0, 10	; Count of bytes to be sent.
	sts		updiUartDataCount_ram, temp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		temp0
	ret

; == Send the Chip Erase key ==
sendChipEraseKey:
	push	temp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendChipEraseKey_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendChipEraseKey_wait
	ldiz	updiUartDataBuf_ram
	ldi		temp0, UPDI_SYNCH
	st		Z+, temp0
	ldi		temp0, 0b111_00_0_00	; Send 64-bit key.
	st		Z+, temp0
	ldi		temp0, 0x65				; Key byte 0.
	st		Z+, temp0
	ldi		temp0, 0x73				; Key byte 1.
	st		Z+, temp0
	ldi		temp0, 0x61				; Key byte 2.
	st		Z+, temp0
	ldi		temp0, 0x72				; Key byte 3.
	st		Z+, temp0
	ldi		temp0, 0x45				; Key byte 4.
	st		Z+, temp0
	ldi		temp0, 0x4D				; Key byte 5.
	st		Z+, temp0
	ldi		temp0, 0x56				; Key byte 6.
	st		Z+, temp0
	ldi		temp0, 0x4E				; Key byte 7.
	st		Z+, temp0
	ldi		temp0, 10	; Count of bytes to be sent.
	sts		updiUartDataCount_ram, temp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		temp0
	ret

; == Issue a System Reset to the target device ==
systemReset:
	push	temp0
	push	ZL
	push	ZH
; Issue the system reset:
; Wait for any previous transmission or reception to be completed:
systemReset_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	systemReset_wait
	ldiz	updiUartDataBuf_ram
	ldi		temp0, UPDI_SYNCH
	st		Z+, temp0
	ldi		temp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	st		Z+, temp0
	ldi		temp0, 0x59					; Normal Reset signature.
	st		Z+, temp0
	ldi		temp0, 3					; Count of bytes to be sent.
	sts		updiUartDataCount_ram, temp0
	rcall	updi_tx
	rcall	updi_idle
; Clear the system reset:
	ldiz	updiUartDataBuf_ram
	ldi		temp0, UPDI_SYNCH
	st		Z+, temp0
	ldi		temp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	st		Z+, temp0
	ldi		temp0, 0x00					; Clear reset.
	st		Z+, temp0
	ldi		temp0, 3					; Count of bytes to be sent.
	sts		updiUartDataCount_ram, temp0
	rcall	updi_tx
	rcall	updi_idle
	pop		ZH
	pop		ZL
	pop		temp0
	ret

; == Send LDCS instruction ==
; Parameters:	temp0, the UPDI register to be loaded.
;				Must be in range 0x00..0x0F.
sendLdcs:
	push	temp1
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendLdcs_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendLdcs_wait
	ldiz	updiUartDataBuf_ram
	ldi		temp1, UPDI_SYNCH
	st		Z+, temp1
	ldi		temp1, 0b100_0_0000				; LDCS instruction.
	or		temp1, temp0
	st		Z+, temp1			
	ldi		temp1, 2
	sts		updiUartDataCount_ram, temp1
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		temp1
	ret

; == Send LDS instruction ==
; Parameters:	temp0, address LSB,
;				temp1, address MSB.
; Original version:
sendLds:
	push	temp2
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendLds_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendLds_wait
	ldiz	updiUartDataBuf_ram
	ldi		temp2, UPDI_SYNCH
	st		Z+, temp2
	ldi		temp2, 0b000_0_01_00			; LDS instruction (word address, 1 byte of data).
	st		Z+, temp2
	st		Z+, temp0						; Address LSB.
	st		Z+, temp1						; Address MSB.
	ldi		temp2, 4
	sts		updiUartDataCount_ram, temp2
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		temp2
	ret

; Version with inter-byte delays of full IDLE cycles:
/* sendLds:
	push	temp2
	ldi		temp2, 1
; Wait for any previous transmission or reception to be completed:
sendLds_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendLds_wait
	sts		updiUartDataCount_ram, temp2
	ldi		temp2, UPDI_SYNCH
	sts		updiUartDataBuf_ram, temp2
	rcall	updi_tx
	rcall	updi_idle
	ldi		temp2, 0b000_0_01_00			; LDS instruction (word address, 1 byte of data).
	sts		updiUartDataBuf_ram, temp2
	rcall	updi_tx
	rcall	updi_idle
	sts		updiUartDataBuf_ram, temp0		; Address LSB.
	rcall	updi_tx
	rcall	updi_idle
	sts		updiUartDataBuf_ram, temp1		; Address MSB.
	rcall	updi_tx
	pop		temp2
	ret */

; == Send STS instruction ==
; Parameters:	temp0, address LSB,
;				temp1, address MSB,
;				temp2, data.
; Return:		temp3, 0: success, 1: timeout, 2: invalid answer.
sendSts:
	push	temp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
sendSts_waitBusy:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	sendSts_waitBusy
; Send instruction code and address:
	ldiz	updiUartDataBuf_ram
	ldi		temp3, UPDI_SYNCH
	st		Z+, temp3						; SYNCH byte.
	ldi		temp3, 0b010_0_01_00
	st		Z+, temp3						; STS instruction (word address, 1 byte of data).
	st		Z+, temp0						; Address LSB.
	st		Z+, temp1						; Address MSB.
	ldi		temp3, 4
	sts		updiUartDataCount_ram, temp3	; Transfer block size.
	rcall	updi_tx
; Read answer byte from Target UPDI:
sendSts_wait0:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	sendSts_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	sendSts_wait0
; updiUartData_ram now contains the received byte.
    lds     temp0, updiUartDataBuf_ram
    cpi		temp0, UPDI_ACK
	brne	sendSts_invAnswer
; Send data byte:
	sts		updiUartDataBuf_ram, temp2		; Data byte to be written.
	ldi		temp3, 1
	sts		updiUartDataCount_ram, temp3	; Transfer block size.
	rcall	updi_tx
; Read answer byte from Target UPDI:
sendSts_wait1:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	sendSts_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	sendSts_wait1
; updiUartData_ram now contains the received byte.
    lds     temp0, updiUartDataBuf_ram
    cpi		temp0, UPDI_ACK
    brne	sendSts_invAnswer
; Success.
	ldi		temp3, 0
	rjmp	sendSts_pop
sendSts_invAnswer:
	ldi		temp3, 2
	rjmp	sendSts_pop
sendSts_timeout:
	ldi		temp3, 1
sendSts_pop:
	pop		ZH
	pop		ZL
	pop		temp0
	ret

; == Wait while NVMCTRL is busy ==
; Return:		temp2, last value of NVMCTRL_STATUS.
;				temp3, 0x00: success, 0x01: timeout.
waitForNvm:
	push	temp0
	push	temp1
waitForNvm_loop:
	ldi		temp0, low(NVMCTRL_STATUS)
	ldi		temp1, high(NVMCTRL_STATUS)
	rcall	sendLds
; Read answer byte from Target UPDI:
waitForNvm_waitRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	waitForNvm_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	waitForNvm_waitRx
; updiUartData_ram now contains the received byte.
    lds     temp2, updiUartDataBuf_ram
	mov		temp1, temp2
	andi	temp1, NVMCTRL_EEBUSY_bm | NVMCTRL_FBUSY_bm
	brne	waitForNvm_loop		; Branch if EEBUSY or FBUSY was set.
	ldi		temp3, 0
	rjmp	waitForNvm_pop
waitForNvm_timeout:
	ldi		temp3, 1
waitForNvm_pop:
	pop		temp1
	pop		temp0
	ret
	