; File:		avr-asm-lib\updi.asm
; Device:	ATtiny3216
; Created:  2024-06-14
; Version:	2024-06-19
; Author:	Johannes Fechner

.equ UPDI_SYNCH = 0x55	; Synchronization byte.
.equ UPDI_ACK = 0x40	; Acknowledge (= "successfully written, ready to accept new transmission" answer) byte.

.equ UPDI_TIMER_UART_PERIOD = ROUND(F_CPU/(1.0*UPDI_BPS))
.equ UPDI_TIMER_ENABLE_TIMEOUT_PERIOD = ROUND(1.0e-3*F_CPU)		; 1ms.
.equ UPDI_TIMER_UART_TIMEOUT_PERIOD = ROUND(1.0*F_CPU/1024.0)	; 1s.
.equ UPDI_RX_TIMEOUT = 5 ; seconds.
.equ UPDI_TX_USE_IDLE_BITS = 0 ; Whether to insert 2 idle bits between transmitted bytes.

; == Routines ==
; === Enable Target UPDI ===
updi_enable:
	push	temp0
; Disable Target UPDI falling-edge interrupt:
	xin		temp0, TARGET_UPDI_PINCTRL
	andi	temp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, temp0
; Drive Target UPDI low for c. 500ns:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
.equ DELAY_LOOP_CYCLES = ROUND(((F_CPU*3.0e-7)-3.0)/2.0)
.if DELAY_LOOP_CYCLES < 1
.error "updi_enable delayLoop has to be replaced by a couple of 'nop' instructions."
.endif
	ldi		temp0, DELAY_LOOP_CYCLES
updi_enable_delayLoop:
	dec		temp0
	brne	updi_enable_delayLoop
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
; Wait for target device to get ready:
; Configure timer for time measuring and timeout detection:
/*	ldi		temp0, low(UPDI_TIMER_ENABLE_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER, temp0
	ldi		temp0, high(UPDI_TIMER_ENABLE_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER+1, temp0
	ldi		temp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, temp0
	ldi		temp0, TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, temp0*/
updi_enable_waitForTarget:
/*	bst		updiFlags, UPDI_FLAGS_TIMEOUT_bp
	brts	updi_enable_afterWait*/
	sbis	TARGET_UPDI_VIN, TARGET_UPDI_bp
	rjmp	updi_enable_waitForTarget
updi_enable_afterWait:
; Stop timer:
/*	xout	TCA0_SINGLE_CTRLA, reg00*/
	pop		temp0
	ret

; === Disable Target UPDI ===
updi_disable:
	push	temp0
; Force guard time delay:
	rcall	updi_idle
	ldi		temp0, UPDI_SYNCH
	sts		updiUartDataBuf_ram, temp0
	ldi		temp0, 0b110_0_0000 | 0x03	; UPDI_CTRLB.
	sts     updiUartDataBuf_ram+1, temp0
	ldi		temp0, 1<<2					; UPDIDIS bit.
	sts     updiUartDataBuf_ram+2, temp0
	ldi		temp0, 3
	sts     updiUartDataCount_ram, temp0
	rcall   updi_tx
	pop		temp0
	ret

; === Wait for c. 25ms (for BREAK) ===
.equ UPDI_BREAK_CPU_CYCLES = ROUND(0.025*F_CPU/10.0)
.if UPDI_BREAK_CPU_CYCLES > 0xFFFF
.error "updi_wait25ms needs a third counter register."
.endif
updi_wait25ms:
	push	tempL
	push	tempH
	ldi		tempL, low(UPDI_BREAK_CPU_CYCLES)
	ldi		tempH, high(UPDI_BREAK_CPU_CYCLES)
updi_wait25ms_loop:
	nop
	nop
	nop
	nop
	nop
	nop
	sbiw	tempH:tempL, 1
	brne	updi_wait25ms_loop
	pop		tempH
	pop		tempL
	ret

; === Send two BREAKs ===
; Duration of a BREAK and the pause between two BREAKs:
updi_break:
; Wait for any previous transmission or reception to be completed:
updi_break_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_break_wait
; Send first BREAK:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rcall	updi_wait25ms
; Pause:
	sbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rcall	updi_wait25ms
; Send second BREAK:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rcall	updi_wait25ms
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rcall	updi_idle
	ret

; === Transmit the byte(s) stored in updiUartDataBuf_ram via UPDI ===
; Parameters:   updiUartDataBuf_ram, the byte(s) to be transmitted (including
;               the preceding SYNCH byte, if necessary); sent in ascending address order.
;               updiUartDataCount_ram, the count of bytes stored in updiUartDataBuf_ram
;               (ranging from 1 to the size of updiUartDataBuf_ram).
updi_tx:
	push	temp0
; Wait for any previous transmission or reception to be completed:
updi_tx_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_tx_wait
; Disable interrupts:
	cli
; Stop UPDI UART timer (previously used for timeout detection):
	xout	TCA0_SINGLE_CTRLA, reg00
; Cancel potential pending timer interrupt:
	xin		temp0, TCA0_SINGLE_INTFLAGS
	ori		temp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTFLAGS, temp0
; Clear TXC, RXC, and timeout flags:
	mov		temp0, updiFlags
	andi	temp0, ~((1<<UPDI_FLAGS_TXC_bp)|(1<<UPDI_FLAGS_RXC_bp)|(1<<UPDI_FLAGS_TIMEOUT_bp))
	mov		updiFlags, temp0
	sei
; Set busy flag:
	set
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
; Initialize bit counter and data byte index:
	sts		updiUartBitCntr_ram, reg00
	sts		updiUartDataIndex_ram, reg00
; Disable Target UPDI edge interrupt:
	xin		temp0, TARGET_UPDI_PINCTRL
	andi	temp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, temp0
; Set Target UPDI pin as low-level output:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	sbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
; Configure, reset and start UPDI UART timer:
	ldi		temp0, low(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER, temp0
	ldi		temp0, high(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER+1, temp0
	ldi		temp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, temp0
	ldi		temp0, TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, temp0
; Re-enable interrupts:
	sei
; Switch on UPDI Tx LED:
	sbi		LED_UPDI_TX_VOUT, LED_UPDI_TX_bp
	pop		temp0
	ret

; === Insert IDLE frame ===
; Wait until current transmission or reception is completed and pause for
; approx. one UPDI UART frame duration.
updi_idle:
	push	tempL
	push	tempH
; Wait for any previous transmission or reception to be completed:
updi_idle_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_idle_wait
; Pause for approx. one UPDI UART bit duration:
.equ UPDI_IDLE_CPU_CYCLES = ROUND(12.0*F_CPU/(UPDI_BPS*10.0))
	ldi		tempL, low(UPDI_IDLE_CPU_CYCLES)
	ldi		tempH, high(UPDI_IDLE_CPU_CYCLES)
updi_idle_delay:
	nop
	nop
	nop
	nop
	nop
	nop
	sbiw	tempL, 1
	brne	updi_idle_delay
	pop		tempH
	pop		tempL
	ret

; == Interrupt Service Routines ==
; === UPDI Falling-Edge ISR ===
updi_fallingEdgeIsr:
	push	temp0
	xin		temp0, CPU_SREG
	push	temp0
; Stop UPDI UART timer (previously used for timeout detection):
	xout	TCA0_SINGLE_CTRLA, reg00
; Clear interrupt flag:
	xin		temp0, TARGET_UPDI_INTFLAGS
	ori		temp0, (1<<TARGET_UPDI_bp)
	xout	TARGET_UPDI_INTFLAGS, temp0
; Clear timer interrupt flag, too:
	xin		temp0, TCA0_SINGLE_INTFLAGS
	ori		temp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTFLAGS, temp0
; Set busy flag:
	set
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
; Disable Target UPDI edge interrupt:
	xin		temp0, TARGET_UPDI_PINCTRL
	andi	temp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, temp0
; Initialize bit counter:
	sts		updiUartBitCntr_ram, reg00
; Configure UPDI UART timer:
	ldi		temp0, low(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER, temp0
	ldi		temp0, high(UPDI_TIMER_UART_PERIOD)
	xout	TCA0_SINGLE_PER+1, temp0
; Load the counter register with half the period so that the received frame is
; sampled in the middle of each bit duration:
	ldi		temp0, low(ROUND(UPDI_TIMER_UART_PERIOD/2.0))
	xout	TCA0_SINGLE_CNT, temp0
	ldi		temp0, high(ROUND(UPDI_TIMER_UART_PERIOD/2.0))
	xout	TCA0_SINGLE_CNT+1, temp0
; Start the timer:
	ldi		temp0, TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, temp0
; Switch on UPDI Rx LED:
	sbi		LED_UPDI_RX_VOUT, LED_UPDI_RX_bp
updi_fallingEdgeIsr_pop:
	pop		temp0
	xout	CPU_SREG, temp0
	pop		temp0
	reti

; === UPDI Timer ISR ===
updi_timerIsr:
	push	temp0
	push	temp1
	push	ZL
	push	ZH
	xin		temp0, CPU_SREG
	push	temp0
; Clear interrupt flag:
	xin		temp0, TCA0_SINGLE_INTFLAGS
	ori		temp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTFLAGS, temp0
; Busy or timer overflow while waiting for reception?
	sbrs    updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp    updi_timerIsr_rxTimeout
; Busy. Increment bit counter and write back:
	lds		temp0, updiUartBitCntr_ram
	inc		temp0
	sts		updiUartBitCntr_ram, temp0
; Transmitting or receiving?
	sbis	TARGET_UPDI_VDIR, TARGET_UPDI_bp
	rjmp	updi_timerIsr_rx					; Target UPDI set as input -> receiving.
; Target UPDI set as output -> currently transmitting.
; temp0 still contains bit counter value.
	cpi		temp0,  1
	breq	updi_timerIsr_txAfterStartBit
	cpi		temp0,  9
	breq	updi_timerIsr_txAfterLastDataBit
	cpi		temp0, 10
	breq	updi_timerIsr_txAfterParityBit
	cpi		temp0, 11
	breq	updi_timerIsr_txAfterParityBit
	cpi		temp0, 12
	breq	updi_timerIsr_txEndOfFrame
.if UPDI_TX_USE_IDLE_BITS
	cpi		temp0, 13
	breq	updi_timerIsr_txIdle0
	cpi		temp0, 14
	brsh	updi_timerIsr_txIdle1
.endif
; (2 <= bit counter <= 8) -> [2nd to 8th] data bit follows.
updi_timerIsr_txInsideData:
; Load data byte and bit mask, set Target UPDI output accordingly:
	ldiz	updiUartDataBuf_ram
	lds		temp0, updiUartDataIndex_ram
	addz	temp0
	ld		temp0, Z
	lds		temp1, updiUartBitMask_ram
	and		temp0, temp1
	breq	updi_timerIsr_txInsideData_0		; Branch if data bit is 0.
	sbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	lds		temp0, updiUartParityCntr_ram
	inc		temp0
	sts		updiUartParityCntr_ram, temp0
	rjmp	updi_timerIsr_txShiftBitMask
updi_timerIsr_txInsideData_0:
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
updi_timerIsr_txShiftBitMask:
; Update bit mask and write back:
	lsl		temp1
	sts		updiUartBitMask_ram, temp1
	rjmp	updi_timerIsr_pop
updi_timerIsr_txAfterStartBit:
; Initialize bit mask and parity counter:
	ldi		temp1, 1
	sts		updiUartBitMask_ram, temp1
	sts		updiUartParityCntr_ram, reg00
	rjmp	updi_timerIsr_txInsideData
updi_timerIsr_txAfterLastDataBit:
; Set Target UPDI output according to parity:
	lds		temp0, updiUartParityCntr_ram
	bst		temp0, 0
	xin		temp0, TARGET_UPDI_VOUT
	bld		temp0, TARGET_UPDI_bp
	xout	TARGET_UPDI_VOUT, temp0
; Increment the byte index:
	lds		temp0, updiUartDataIndex_ram
	inc		temp0
	sts		updiUartDataIndex_ram, temp0
	rjmp	updi_timerIsr_pop
updi_timerIsr_txAfterParityBit:
; Stop bit (high) follows:
	sbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_txEndOfFrame:
; After 2nd=last stop bit -> end of frame.
; Check whether end of transmission is reached:
	lds		temp0, updiUartDataIndex_ram
	lds		temp1, updiUartDataCount_ram
	cp		temp0, temp1
	brsh	updi_timerIsr_txEot		; Branch if EOT reached.
.if UPDI_TX_USE_IDLE_BITS
; Else, insert inter-byte delay of 2 idle bits.
updi_timerIsr_txIdle0:
	rjmp	updi_timerIsr_pop
updi_timerIsr_txIdle1:
.endif
; Reset bit counter and drive UPDI low for start bit:
	sts		updiUartBitCntr_ram, reg00
	cbi		TARGET_UPDI_VOUT, TARGET_UPDI_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_txEot:
; End of transmission reached.
; Stop timer:
	xout	TCA0_SINGLE_CTRLA, reg00
; Set Target UPDI pin as input:
	cbi		TARGET_UPDI_VDIR, TARGET_UPDI_bp
; Clear busy flag, set TXC flag:
	clt
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
	set
	bld		updiFlags, UPDI_FLAGS_TXC_bp
; Enable Target UPDI edge interrupt:
	xin		temp0, TARGET_UPDI_PINCTRL
	ori		temp0, PORT_ISC_FALLING_gc
	xout	TARGET_UPDI_PINCTRL, temp0
; Clear data byte index:
	sts		updiUartDataIndex_ram, reg00
; Configure timer for RX-timeout detection:
	sts		updiUartTimeoutCntr_ram, reg00
	ldi		temp0, low(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER, temp0
	ldi		temp0, high(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER+1, temp0
	ldi		temp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, temp0
	ldi		temp0, TCA_SINGLE_CLKSEL_DIV1024_gc | TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, temp0
; Switch off UPDI Tx LED:
	cbi		LED_UPDI_TX_VOUT, LED_UPDI_TX_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_rxTimeout:
	lds		temp0, updiUartTimeoutCntr_ram
	inc		temp0
	sts		updiUartTimeoutCntr_ram, temp0
	cpi		temp0, UPDI_RX_TIMEOUT
	brlo	updi_timerIsr_noTimeout
; Timeout reached.
; Disable Target UPDI edge interrupt:
	xin		temp0, TARGET_UPDI_PINCTRL
	andi	temp0, ~PORT_ISC_gm
	xout	TARGET_UPDI_PINCTRL, temp0
; Disable potential pending interrupt:
	xin		temp0, TARGET_UPDI_INTFLAGS
	ori		temp0, (1<<TARGET_UPDI_bp)
	xout	TARGET_UPDI_INTFLAGS, temp0
; Stop timer:
	xout	TCA0_SINGLE_CTRLA, reg00
; Set timeout flag:
	set
	bld     updiFlags, UPDI_FLAGS_TIMEOUT_bp
updi_timerIsr_noTimeout:
	rjmp    updi_timerIsr_pop
updi_timerIsr_rx:
; temp0 still contains bit counter value.
	cpi		temp0, 1
	breq	updi_timerIsr_rxStart
	cpi		temp0, 10
	breq	updi_timerIsr_rxParity
	cpi		temp0, 11
	breq	updi_timerIsr_rx1stStop
	cpi		temp0, 12
	brsh	updi_timerIsr_rx2ndStop
; (2 <= bit counter <= 9) -> data bit.
	ldiz	updiUartDataBuf_ram
	lds		temp0, updiUartDataIndex_ram
; Make sure index is in allowed range:
	cpi		temp0, UPDI_UART_BUFLEN
	brsh	updi_timerIsr_bufovf
; Load the appropriate data byte from buffer:
	addz	temp0
	ld		temp0, Z
	lds		temp1, updiUartBitMask_ram
; If UPDI is high, set the appropriate bit:
	sbic	TARGET_UPDI_VIN, TARGET_UPDI_bp
	or		temp0, temp1
	st		Z, temp0
; Left-shift the bit mask and write back:
	lsl		temp1
	sts		updiUartBitMask_ram, temp1
	rjmp	updi_timerIsr_pop
updi_timerIsr_bufovf:
	set
	bld		updiFlags, UPDI_FLAGS_BUFOVF_bp
	rjmp	updi_timerIsr_pop
updi_timerIsr_rxStart:
; Initialize the bit mask and clear the data byte:
	ldi		temp0, 1
	sts		updiUartBitMask_ram, temp0
	ldiz	updiUartDataBuf_ram
	lds		temp0, updiUartDataIndex_ram
; Make sure index is in allowed range:
	cpi		temp0, UPDI_UART_BUFLEN
	brsh	updi_timerIsr_bufovf
	addz	temp0
	st		Z, reg00
	rjmp	updi_timerIsr_pop
updi_timerIsr_rxParity:
; ##### To be done: test for correct parity. #####
; Increment byte index:
	lds		temp0, updiUartDataIndex_ram
	inc		temp0
	sts		updiUartDataIndex_ram, temp0
updi_timerIsr_rx1stStop:
; ##### To be done: test for high level. #####
	rjmp	updi_timerIsr_pop
updi_timerIsr_rx2ndStop:
; End of frame reached.
; ##### To be done: test for high level. #####
; Stop timer:
	xout	TCA0_SINGLE_CTRLA, reg00
; Set UPDI Frame Received flag:
	set
	bld		updiFlags, UPDI_FLAGS_RXC_bp
; Clear busy flag:
    clt
	bld     updiFlags, UPDI_FLAGS_BUSY_bp
; Enable Target UPDI edge interrupt:
	xin		temp0, TARGET_UPDI_PINCTRL
	ori		temp0, PORT_ISC_FALLING_gc
	xout	TARGET_UPDI_PINCTRL, temp0
; Configure timer for RX-timeout detection:
	sts		updiUartTimeoutCntr_ram, reg00
	ldi		temp0, low(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER, temp0
	ldi		temp0, high(UPDI_TIMER_UART_TIMEOUT_PERIOD)
	xout	TCA0_SINGLE_PER+1, temp0
	ldi		temp0, TCA_SINGLE_CMD_RESTART_gc
	xout	TCA0_SINGLE_CTRLESET, temp0
	ldi		temp0, TCA_SINGLE_CLKSEL_DIV1024_gc | TCA_SINGLE_ENABLE_bm
	xout	TCA0_SINGLE_CTRLA, temp0
; Switch off UPDI Rx LED:
	cbi		LED_UPDI_RX_VOUT, LED_UPDI_RX_bp
updi_timerIsr_pop:
	pop		temp0
	xout	CPU_SREG, temp0
	pop		ZH
	pop		ZL
	pop		temp1
	pop		temp0
	reti
